'  Sunflower Generation Program
'  Jeff Bowermaster
'  Splat! Graphics
'  8/16/92

'  Generates a "Sunflower Seed Head" collection of spheres
'  using a collision model similar to some stuff I read in the
'  1992 SIGGRAPH proceedings.

DECLARE SUB rotate (d, e, f)

TYPE vector
  x AS SINGLE
  y AS SINGLE
  z AS SINGLE
END TYPE

DIM red(16), green(16), blue(16), v(5000) AS vector
COMMON SHARED rad, xrotate, yrotate, zrotate
SCREEN 12

FOR y = 1 TO 4
	FOR x = 1 TO 4
		 colornum = x + ((y - 1) * 4) - 1
		 READ red(colornum), green(colornum), blue(colornum)
		 col = 65536 * blue(colornum) + 256 * green(colornum) + red(colornum)
		 PALETTE colornum, col
		 COLOR colornum
	NEXT x
NEXT y


'Rainbow Palette

DATA 0,  0,  0
DATA 32,  0,  0
DATA 42,  0,  0
DATA 58, 16,  0
DATA 63, 32,  0
DATA 58, 56,  0
DATA 16, 42,  0
DATA  0, 30, 36
DATA  0, 20, 40
DATA  0, 10, 48
DATA  0,  0, 63
DATA 20,  0, 53
DATA 23,  0, 29
DATA 19,  7, 17
DATA 50, 40, 45
DATA 63, 63, 63

pi = 3.1415926536#
rad = pi / 180
cr = 1
min = .02
min2 = min * 2

WINDOW (-3.2, -2.4)-(3.2, 2.4)

zrotate = 137.5     'evolution angle

	n = 1
	v(n).x = 0
	v(n).y = 0
	v(n).z = 0

CIRCLE (v(n).x, v(n).y), min

	n = 2
	v(n).x = 2 * min
	v(n).y = 0
	v(n).z = cr - (cr ^ 2 - (v(n).x ^ 2 + v(n).y ^ 2)) ^ .5

CIRCLE (v(n).x, v(n).y), min

	n = 3

DO WHILE n < 1000
	LOCATE 1, 1: PRINT n
	v(n) = v(n - 1)   'make the last point the basis for the next one

	min2 = min * 2 + (EXP(n / 250) / 100)   ' rate of growth

	CALL rotate(v(n).x, v(n).y, v(n).z)

	again = 0
retry:
	again = again + 1     ' iterative counter (3 seems to work OK)

	close1 = 2            ' arbitrary - larger than expected closeness
	close2 = 2

	yes = 0               ' did we find two overlapping circles?

	FOR m = n - 1 TO 1 STEP -1
		' look at every circle we have so far, and compare it to this one

		d = ((v(m).x - v(n).x) ^ 2 + (v(m).y - v(n).y) ^ 2 + (v(m).z - v(n).z) ^ 2) ^ .5

		' find the two closest circles to it

		IF close1 > d THEN
			close2 = close1         ' this sets the next to closest one
			cn2 = cn1               ' requires the closest one follow the next
										  ' closest one ...
			close1 = d              ' this is the closest one
			cn1 = m
		ELSE
			IF close1 < d AND close2 > d THEN
			  close2 = d
			  cn2 = m               ' this abomination catches when the
			END IF                  ' next to closest one is encountered
		END IF                     ' after the closest one during the scan

	  IF d < min2 THEN yes = 1    ' yes, there is overlap
  NEXT m

  IF yes = 1 THEN
	  v(n).x = (v(cn1).x + v(cn2).x) / 2   ' set the circle between the
	  v(n).y = (v(cn1).y + v(cn2).y) / 2   ' two circles it's closest to
	  v(n).z = cr - ABS(cr ^ 2 - (v(n).x ^ 2 + v(n).y ^ 2)) ^ .5

	  'the line connecting the two closest circles

	 'a line perpendicular would then be -y, x

	  slx = v(cn1).x - v(cn2).x
	  sly = v(cn1).y - v(cn2).y

	  'once a point and a slope are determined, the direction away from the
	  'center (0,0) along that line must be determined.  In all cases, we want
	  'to move away from 0,0.  A simple test sets the required direction.
		
		a = v(n).x
		b = v(n).y
		d1 = a ^ 2 + b ^ 2  ' - 0 ^ 2 implied
		a = a - .01 * sly
		b = b + .01 * slx
		d2 = a ^ 2 + b ^ 2
		IF d1 > d2 THEN s = -1 ELSE s = 1
		r = d1 ^ .5   ' r is used to reduce the search step size as we move
						  ' away from the center
	END IF

  ' distance from both centers

	d1 = ((v(cn1).x - v(n).x) ^ 2 + (v(cn1).y - v(n).y) ^ 2 + (v(cn1).z - v(n).z) ^ 2) ^ .5
	d2 = ((v(cn2).x - v(n).x) ^ 2 + (v(cn2).y - v(n).y) ^ 2 + (v(cn2).z - v(n).z) ^ 2) ^ .5

	count = 0
	DO WHILE d1 < min2 AND d2 < min2
		count = count + 1   ' accelerate the solution (it pauses unless you do this)
		v(n).x = v(n).x - .01 * sly * s * count / (25 * r)
		v(n).y = v(n).y + .01 * slx * s * count / (25 * r)
		v(n).z = cr - ABS(cr ^ 2 - (v(n).x ^ 2 + v(n).y ^ 2)) ^ .5

		d1 = ((v(cn1).x - v(n).x) ^ 2 + (v(cn1).y - v(n).y) ^ 2 + (v(cn1).z - v(n).z) ^ 2) ^ .5
		d2 = ((v(cn2).x - v(n).x) ^ 2 + (v(cn2).y - v(n).y) ^ 2 + (v(cn2).z - v(n).z) ^ 2) ^ .5
	LOOP

	'This solution may not be a global one.  We may have moved the
	'circle into another one already that initially wasn't close. Try again
	IF again < 1 THEN GOTO retry

	CIRCLE (v(n).x, v(n).z), min2 / 2, (n MOD 12) + 1

'  triangles, but really bad ones
'   LINE (v(n).x, v(n).y)-(v(cn1).x, v(cn1).y)
'   LINE (v(n).x, v(n).y)-(v(cn2).x, v(cn2).y)
'   LINE (v(cn1).x, v(cn1).y)-(v(cn2).x, v(cn2).y)


	'DO WHILE INKEY$ = "": LOOP
	n = n + 1
LOOP

OPEN "flower.vo" FOR OUTPUT AS #1
CLS
FOR n = 1 TO 999
	min2 = min * 2 + (EXP(n / 250) / 100)
	CIRCLE (v(n).x, v(n).y), min2 / 2
	PRINT #1, USING "sphere = { center = ##.######  ##.######  ##.######; radius =  #.######; }"; v(n).x, v(n).y, v(n).z, min2 / 2
NEXT n
CLOSE #1

SUB rotate (d, e, f)

	 x0 = d
	 y0 = e
	 z0 = f

	 x1 = x0
	 y1 = y0 * COS(xrotate * rad) - z0 * SIN(xrotate * rad)
	 z1 = y0 * SIN(xrotate * rad) + z0 * COS(xrotate * rad)

	 x2 = z1 * SIN(yrotate * rad) + x1 * COS(yrotate * rad)
	 y2 = y1
	 z2 = z1 * COS(yrotate * rad) - x1 * SIN(yrotate * rad)

	 x3 = x2 * COS(zrotate * rad) - y2 * SIN(zrotate * rad)
	 y3 = x2 * SIN(zrotate * rad) + y2 * COS(zrotate * rad)
	 z3 = z2

	 d = x3
	 e = y3
	 f = z3

END SUB

